const  _    = require('lodash');
const fetch = require('isomorphic-fetch');
const {
  Schema,
  arrayOf,
  normalize
} = require('normalizr');
const {
  apiHandler,
  parseJSON,
  makeQueryString
} = require('./response');
const {
  API_ROOT
} = require('./constants');

/* Define schemas used for Api responses
 */
const schemaOpts = { idAttribute: '_id' };
const sessionSchema = new Schema('sessions', schemaOpts);
const recipeSchema = new Schema('recipes', schemaOpts);
const mealSchema = new Schema('meals', schemaOpts);
const userSchema = new Schema('users', schemaOpts);
const foodSchema = new Schema('foods', schemaOpts);
const ingredientFoodsSchema = new Schema('ingredientFoods', schemaOpts);
const ingredientRecipesSchema = new Schema('ingredientRecipes', schemaOpts);

sessionSchema.define({
  user: userSchema
});

foodSchema.define({
  author: userSchema
});

ingredientFoodsSchema.define({
  food: foodSchema
});

recipeSchema.define({
  author: userSchema,
  ingredientFoods: arrayOf(ingredientFoodsSchema)
});

ingredientRecipesSchema.define({
  recipe: recipeSchema
});

mealSchema.define({
  author: userSchema,
  ingredientFoods: arrayOf(ingredientFoodsSchema),
  ingredientRecipes: arrayOf(ingredientRecipesSchema)
});

/* Export available schemas corresponding for every
 * Api response (used by /actions).
 */
const SCHEMAS = {
  SESSION: sessionSchema,

  USER: userSchema,
  USERS: arrayOf(userSchema),

  MEAL: mealSchema,
  MEALS: arrayOf(mealSchema),

  RECIPE: recipeSchema,
  RECIPES: arrayOf(recipeSchema),

  FOOD: foodSchema,
  FOODS: arrayOf(foodSchema)
};

/* Special symbol that signals this middleware to be
 * called in the redux middleware chain (used by /actions).
 */
const CALL_API = 'Call API';

function assertValidCall({ query, body, method, endpoint, schema, types }) {
  if (query && body) {
    throw new Error('Do not call with both query and body');
  }
  if (!method) {
    throw new Error('Specify a method for api call.');
  }
  if (['GET', 'POST', 'PUT', 'DELETE'].indexOf(method) === -1) {
    throw new Error('Method must be one of {GET, POST, PUT, DELETE}');
  }
  if (typeof endpoint !== 'string') {
    throw new Error('Specify a string endpoint URL.');
  }
  if (!schema) {
    throw new Error('Specify one of the exported Schemas.');
  }
  if (!Array.isArray(types) || types.length !== 4) {
    throw new Error('Expected an array of four action types.');
  }
  if (!types.every(type => typeof type === 'string')) {
    throw new Error('Expected action types to be strings.');
  }
}

/* Middleware that interprets calls with CALL_API.
 */
const apiFn = store => next => action => {
  const callAPI = action[CALL_API];
  if (typeof callAPI === 'undefined') {
    return next(action);
  }

  const { token, query, body, method, endpoint, schema, types } = callAPI;

  assertValidCall(callAPI);

  let url = `${API_ROOT}/${endpoint}`;

  const [reqType, okType, cancelType, errType] = types;

  const options = {
    credentials: 'include',
    method: method,
    headers: {
      'Accept': 'application/json',
      'Content-Type': 'application/json'
    }
  };

  if (token) {
    options.headers.Authorization = token;
  }

  if (body) {
    options.body = JSON.stringify(body);
  }

  if (query && !_.isEmpty(query)) {
    url = `${url}?${makeQueryString(query)}`;
  }

  next({ type: reqType });
  return fetch(url, options)
    .then(parseJSON)
    .then(apiHandler(
      json => {
        next({
          type: okType,
          response: normalize(json, schema)
        });
      },
      (error, mongoose) => {
        next({ type: cancelType });
        next({
          type: errType,
          error,
          mongoose
        });
      }
    ));
};

module.exports = {
  CALL_API,
  SCHEMAS,
  apiFn
};
